#include <Adafruit_NeoPixel.h>

// This version uses text image file and bat and santa image files with grayscale only (not r,g,b) to save program memory.
// More text can be written or other animations could be added.
// sketch_rgb_to_grayscale.ino was used to convert the r,g,b files to grayscale.
// This version uses separate data pins for each panel, so that less memory is needed by the Neopixel library, and Arduino Uno can be used.
// This program displays a series of scrolling images and effects on three 16x16 LED arrays.
// First, scrolling text "Happy Halloween!"
// Then, a snow effect
// Then, a rising full moon with flying santa.
// Then, a rising star
// Then, star rises above blinking Christmas tree lights

#define PIN      6  // change to whatever pin you are using to send data to the LED array.  This is now replaced by the array panelPin[]
#define N_LEDS 256  // number of LEDs in one panel.  Originally this was the total number of LEDs in all panels.
#define imageRows 16  // the size of the image that is going to be displayed.  Height normally 16.
#define merryCols 149  // Change this number depending on the width of your scrolling image.
#define noelCols 102  // Change this number depending on the width of your scrolling image.
// number of rows and columns in each LED panel
#define numRows 16
#define numCols 16
#define moonRows 48
#define moonCols 48
#define presentRows 18
#define presentCols 48
#define starRows 16
#define starCols 16
#define santaRows 16
#define santaCols 70
#define snowDuration 200
#define snowCooldown 64
#define brightness 70
#define count1max 2
#define count0max 4
#define numColors 8  // number of colors of lights on Christmas tree
#define treeHeight 96
#define treeDuration 48

Adafruit_NeoPixel strip = Adafruit_NeoPixel(N_LEDS, PIN, NEO_GRB + NEO_KHZ800);

// int numPanels = N_LEDS/numRows/numCols;
int numPanels = 3; 
int panelPin[3] = {5,6,7};  // Panel 0 connected to pin 5, panel 1 to pin 6, panel 2 to pin 7.
int displayCols = numCols*numPanels;
int offset = displayCols;
int offset2 = 0;
int imageCols[2] = {149,102};  // columns in Merry Christmas and Joyeux Noel messages
int count = 0;
int count0 = 0;
int count1 = 0;
int pixel = 0;
int imageStartCol = 0;
int imageRow = 0;
int imageRow2 = 0;
int starRow2 = 0;
int presentRow2 = 0;
int imageCol = 0;
int starCol = 0;
int starStartCol = 0;
int santaRow = 0;
int santaCol = 0;
int snowValue = 0;
int snowValue2 = 0;
int treeWidth = 0;
int r=0;
int g=0;
int b=0;
int red=0;
int green=0;
int blue=0;
int hueIndex = 0;
int hue1 = 0;
int hue2 = 0;
int x=0;
int y=0;
int y2=0;
int panel = 0;
int flag = 1;  //  alternate between santa and bat, and change rainbow
int value = 0;
long randValue;

//define the width and height of the screen and the buffers
//#define screenHeight numRows
int w = displayCols;
int h = numRows;
int h1 = h-1;
int w2 = w/2;
byte snow[17][48];  // bit 0 indicates presence of a small snowflake, bit 1 a larger snowflake, bit 2 largest snowflake
const uint8_t palette[8] PROGMEM = 
{0, 8, 16, 24, 50, 60, 80, 100};

// colors for Christmas tree lights
byte colors[8][3] = 
{{0, 0, 0}, {0, 0, 255}, {0, 255, 0}, {255, 0, 0}, {0, 255, 255}, 
{255, 255, 0}, {255, 0, 255}, {255, 80, 80}};


// the image data is stored in program memory (flash memory), because there is not enough RAM in an Arduino to store the whole image.
const uint8_t mc[imageRows][merryCols] PROGMEM = 
#include "merryg.h"   // Happy Halloween written in white with black background
;
const uint8_t noel[imageRows][noelCols] PROGMEM = 
#include "noelg.h"   // Joyeux Noel written in white with black background
;
const uint8_t santa[santaRows][santaCols] PROGMEM = 
#include "santag.h"   // black santa silhouette on white background
;
const uint8_t moon[moonRows][moonCols][3] PROGMEM = 
#include "full_moon_darker_48x48.h"   // full moon image on black background.
;
const uint8_t star[starRows][starCols][3] PROGMEM = 
#include "star2_16x16.h"   // gold star on black background
;
const uint8_t presents[presentRows][presentCols][3] PROGMEM = 
#include "presents_zeros_48x18.h"   // presents on black background
;

void setup() {
  strip.begin();
  for (panel = 0; panel < numPanels; panel++) {
    strip.setPin(panelPin[panel]);
    strip.show();         // Turn OFF all pixels ASAP
  }
  strip.setBrightness(brightness); // Set BRIGHTNESS

}

void loop() {
  
// scrolling text followed by snow effect

  // make sure the snow buffer is zero in the beginning
  for(y = 0; y < h; y++) {
    for(x = 0; x < w; x++) {
      snow[y][x] = 0;
    }
  }
  for (offset = displayCols; offset > -imageCols[flag-1]-snowDuration; offset--) {
    offset2 = offset+imageCols[flag-1];  // this quantity is used several times, store it to save computation time
     // compute snow effect first.  snow effect only applies for columns greater than (imageCols-offset) or 0, and less than displayCols
    //do the snow calculations for every pixel, from bottom to top
      for(y = h1; y > 0; y--) {  // move layer 2 snowflakes every loop
        for(x = max(offset2,0); x < w; x++) {
          value = bitRead(snow[y-1][x],2);
          bitWrite(snow[y][x],2,value); 
        }
      }
      for(x = max(offset2,0); x < w; x++) {
        bitClear(snow[0][x],2);
      }
      if (offset > -imageCols[flag-1]-snowDuration+snowCooldown) {
        for(x = max(offset2,0); x < w; x++) {  // start new snowflakes
          randValue = random(60);
          if (randValue==1L) {
            snow[0][x] = 4;
          }
        }
      }
    count1++;
    if (count1 == count1max) {   // update plane 1 only every second cycle
      count1 = 0;
      for(y = h1; y > 0; y--) {
        for(x = max(offset2,0); x < w; x++) {
          value = bitRead(snow[y-1][x],1);
          bitWrite(snow[y][x],1,value); 
        }
      }
      for(x = max(offset2,0); x < w; x++) {
        bitClear(snow[0][x],1);
      }
      if (offset > -imageCols[flag-1]-snowDuration+snowCooldown) {
        for(x = max(offset2,0); x < w; x++) {  // start new snowflakes
          randValue = random(30);
          if (randValue==1L) {
            bitSet(snow[0][x],1);
          }
        }
      }
    }
    count0++;
    if (count0 == count0max) {   // update plane 0 only every fourth cycle
      count0 = 0;
      for(y = h1; y > 0; y--) {
        for(x = max(offset2,0); x < w; x++) {
          value = bitRead(snow[y-1][x],0);
          bitWrite(snow[y][x],0,value); 
        }
      }
      for(x = max(offset2,0); x < w; x++) {
        bitClear(snow[0][x],0);
      }
      if (offset > -imageCols[flag-1]-snowDuration+snowCooldown) {
        for(x = max(offset2,0); x < w; x++) {  // start new snowflakes
          randValue = random(15);
          if (randValue==1L) {
            bitSet(snow[0][x],0);
          }
        }
      }
    }

    // now update the display, one pixel at a time
  
   for(panel=0; panel<numPanels; panel++) {
    pixel = 0;
    imageStartCol = panel*numCols-offset;
    imageRow = numRows - 1;
    for (y=0; y<(numRows>>1); y++) {
      hueIndex = ((offset2)<<2) + (imageRow<<4)*flag;
      if (hueIndex>=384) {
        hueIndex -= 384;
      }
      if (hueIndex>=384) {
        hueIndex -= 384;
      }
      if (hueIndex>=384) {
        hueIndex -= 384;
      }
// create rainbow hues that transition from red (hueIndex 0) to green (hueIndex 128) to white (hueIndex 256) to red (hueIndex 384)
      hue1 = max(hueIndex,128);
      hue2 = max(hueIndex,256);
//      red = 128-hueIndex + 2*(hue1-128) - (hue2-256);  // red, green, and blue only go to 128 to avoid overflow when multiplying by 255
      red = ((hue1)<<1) - hueIndex  - hue2 + 128;  // simplify
//      green = hueIndex - (hue1-128) - (hue2-256);
      green = hueIndex - hue1 - hue2 + 384;  // simplify
//      blue = hue1-128 - ((hue2-256)<<1);
      blue = hue1+384 - ((hue2)<<1);  // simplify
      for (x=0; x<numCols; x++) {
        imageCol = imageStartCol + x;
        if ((imageCol < 0)) {  // Draw black (text not present yet)
          r = 0;
          g = 0;
          b = 0;
        }
        else if ((imageCol >= imageCols[flag-1])) {  // draw snow
          snowValue = int(snow[imageRow][x+panel*numCols]);
            r = int(pgm_read_byte_near(&palette[snowValue]));
            g = r;
            b = r;
        }
        else {  // draw rainbow text
          if (flag == 1) {
            r = (int(pgm_read_byte_near(&mc[imageRow][imageCol]))*red)>>7;  // use the bitshift function to divide by 128
            g = (int(pgm_read_byte_near(&mc[imageRow][imageCol]))*green)>>7;
            b = (int(pgm_read_byte_near(&mc[imageRow][imageCol]))*blue)>>7;
          }
          else {
            r = (int(pgm_read_byte_near(&noel[imageRow][imageCol]))*red)>>7;  // use the bitshift function to divide by 128
            g = (int(pgm_read_byte_near(&noel[imageRow][imageCol]))*green)>>7;
            b = (int(pgm_read_byte_near(&noel[imageRow][imageCol]))*blue)>>7;
          }
        }
        drawPixel(strip.Color(r, g, b)); // Draw pixel
        pixel++;
      }
      imageRow--;
      hueIndex = (offset2<<2) + (imageRow<<4)*flag;
      if (hueIndex>=384) {
        hueIndex -= 384;
      }
      if (hueIndex>=384) {
        hueIndex -= 384;
      }
      if (hueIndex>=384) {
        hueIndex -= 384;
      }
      hue1 = max(hueIndex,128);
      hue2 = max(hueIndex,256);
//      red = 128-hueIndex + 2*(hue1-128) - (hue2-256);  // red, green, and blue only go to 128 to avoid overflow when multiplying by 255
      red = ((hue1)<<1) - hueIndex  - hue2 + 128;  // simplify
//      green = hueIndex - (hue1-128) - (hue2-256);
      green = hueIndex - hue1 - hue2 + 384;  // simplify
//      blue = hue1-128 - ((hue2-256)<<1);
      blue = hue1+384 - ((hue2)<<1);  // simplify
      for (x=numCols-1; x>=0; x--) {
        imageCol = imageStartCol + x;
        if ((imageCol < 0)) {
          r = 0;
          g = 0;
          b = 0;
        }
        else if ((imageCol >= imageCols[flag-1])) {  // draw snow
          snowValue = int(snow[imageRow][x+panel*numCols]);
          r = int(pgm_read_byte_near(&palette[snowValue]));
          g = r;
          b = r;
        }
        else {  // draw rainbow text
          if (flag == 1) {
            r = (int(pgm_read_byte_near(&mc[imageRow][imageCol]))*red)>>7;  // use the bitshift function to divide by 128
            g = (int(pgm_read_byte_near(&mc[imageRow][imageCol]))*green)>>7;
            b = (int(pgm_read_byte_near(&mc[imageRow][imageCol]))*blue)>>7;
          }
          else {
            r = (int(pgm_read_byte_near(&noel[imageRow][imageCol]))*red)>>7;  // use the bitshift function to divide by 128
            g = (int(pgm_read_byte_near(&noel[imageRow][imageCol]))*green)>>7;
            b = (int(pgm_read_byte_near(&noel[imageRow][imageCol]))*blue)>>7;
          }
        }
        drawPixel(strip.Color(r, g, b)); // Draw image pixel
        pixel++;
      }
      imageRow--;
    }
    strip.setPin(panelPin[panel]);
    strip.show();
   }
  //  delay(0);  // you can change this number to change the scrolling speed.  Smaller number will be faster scrolling.
  }


// rising moon with santa passing in front

  for (count = -numRows<<1; count < (moonRows<<1)-1; count++) {
    offset = (count>>1);
    offset2 = ((count+1)>>1);
    // now update the display, one pixel at a time
  
    for(panel=0; panel<numPanels; panel++) {
      pixel = 0;
      imageStartCol = panel*numCols;
      imageRow = numRows - 1 + offset;
      imageRow2 = numRows - 1 + offset2;
      santaRow = numRows - 1;
      for (y=0; y<numRows>>1; y++) {
        for (x=0; x<numCols; x++) {
          imageCol = imageStartCol + x;
          santaCol = imageCol - count + santaCols - numRows - 13;
          if ((imageRow < 0) || (imageRow2 >= moonRows)) {  // Draw black (moon not present yet)
            r = 0;
            g = 0;
            b = 0;
          }
          else if ((santaCol >= 0) && (santaCol < santaCols)) {  // draw moon with santa
            r = int(((long(pgm_read_byte_near(&moon[imageRow][imageCol][0]))+long(pgm_read_byte_near(&moon[imageRow2][imageCol][0])))>>1)*long(pgm_read_byte_near(&santa[santaRow][santaCol]))>>8);
            g = int(((long(pgm_read_byte_near(&moon[imageRow][imageCol][1]))+long(pgm_read_byte_near(&moon[imageRow2][imageCol][1])))>>1)*long(pgm_read_byte_near(&santa[santaRow][santaCol]))>>8);
            b = int(((long(pgm_read_byte_near(&moon[imageRow][imageCol][2]))+long(pgm_read_byte_near(&moon[imageRow2][imageCol][2])))>>1)*long(pgm_read_byte_near(&santa[santaRow][santaCol]))>>8);
          }
          else {  // draw moon
            r = (int(pgm_read_byte_near(&moon[imageRow][imageCol][0]))+int(pgm_read_byte_near(&moon[imageRow2][imageCol][0])))>>1;
            g = (int(pgm_read_byte_near(&moon[imageRow][imageCol][1]))+int(pgm_read_byte_near(&moon[imageRow2][imageCol][1])))>>1;
            b = (int(pgm_read_byte_near(&moon[imageRow][imageCol][2]))+int(pgm_read_byte_near(&moon[imageRow2][imageCol][2])))>>1;
          }
          drawPixel(strip.Color(r, g, b)); // Draw pixel
          pixel++;
        }
        imageRow--;
        imageRow2--;
        santaRow--;
        for (x=numCols-1; x>=0; x--) {
          imageCol = imageStartCol + x;
          santaCol = imageCol - count + santaCols - numRows - 13;
          if ((imageRow < 0) || (imageRow2 >= moonRows)) {
            r = 0;
            g = 0;
            b = 0;
          }
          else if ((santaCol >= 0) && (santaCol < santaCols)) {  // draw moon with santa
            r = int(((long(pgm_read_byte_near(&moon[imageRow][imageCol][0]))+long(pgm_read_byte_near(&moon[imageRow2][imageCol][0])))>>1)*long(pgm_read_byte_near(&santa[santaRow][santaCol]))>>8);
            g = int(((long(pgm_read_byte_near(&moon[imageRow][imageCol][1]))+long(pgm_read_byte_near(&moon[imageRow2][imageCol][1])))>>1)*long(pgm_read_byte_near(&santa[santaRow][santaCol]))>>8);
            b = int(((long(pgm_read_byte_near(&moon[imageRow][imageCol][2]))+long(pgm_read_byte_near(&moon[imageRow2][imageCol][2])))>>1)*long(pgm_read_byte_near(&santa[santaRow][santaCol]))>>8);
          }
          else {  // draw moon
            r = (int(pgm_read_byte_near(&moon[imageRow][imageCol][0]))+int(pgm_read_byte_near(&moon[imageRow2][imageCol][0])))>>1;
            g = (int(pgm_read_byte_near(&moon[imageRow][imageCol][1]))+int(pgm_read_byte_near(&moon[imageRow2][imageCol][1])))>>1;
            b = (int(pgm_read_byte_near(&moon[imageRow][imageCol][2]))+int(pgm_read_byte_near(&moon[imageRow2][imageCol][2])))>>1;
          }
          drawPixel(strip.Color(r, g, b)); // Draw image pixel
          pixel++;
        }
        imageRow--;
        imageRow2--;
        santaRow--;
      }
      strip.setPin(panelPin[panel]);
      strip.show();
    }
  }

  // rising star, followed by Christmas tree lights, then presents under the tree

  // make sure the snow buffer is zero in the beginning
  for(y = 0; y <= h; y++) {
    for(x = 0; x < w; x++) {
      snow[y][x] = 0;
    }
  }

  for (count = -numRows<<1; count < (starRows+treeDuration+presentRows)*2-1; count++) {
    offset = (count>>1);
    offset2 = ((count+1)>>1);
    treeWidth = offset/2+1;

    // move tree lights every second count.  Re-use snow array for the tree lights effect.
    // do the tree lights calculations for every pixel, from top to bottom
    if (offset!=offset2) {
      for(y = h; y > 0; y--) {
        for(x = 0; x < w; x++) {
          snow[y][x] = snow[y-1][x]; 
        }
      }
      for(x = 0; x < w; x++) {   // generate new tree lights in bottom row
        snow[0][x] = 0;
        if ((x >= w2-treeWidth)&&(x < w2+treeWidth)) {
          randValue = random(30);

          if ((randValue > 7)||(offset>treeDuration)) randValue = 0;
          snow[0][x] = byte(randValue);
        }
      }
    }

    // now update the display, one pixel at a time
  
    for(panel=0; panel<numPanels; panel++) {
      pixel = 0;
      starStartCol = (panel-1)*numCols;
      imageStartCol = panel*numCols;
      imageRow = numRows + offset;
      imageRow2 = numRows + offset2;
      for (y=0; y<numRows>>1; y++) {
        y2 = y*2;
        for (x=0; x<numCols; x++) {
          imageCol = imageStartCol + x;
          starCol = starStartCol + x;
          if ((imageRow < 0)||(imageRow >= starRows+treeDuration+presentRows)) {  // Draw black (star not present yet, or past presents)
            r = 0;
            g = 0;
            b = 0;
          }
          else if (imageRow >= starRows+treeDuration) {  // draw presents
            presentRow2 = min(imageRow2,presentRows+starRows+treeDuration-1);
            r = (int(pgm_read_byte_near(&presents[imageRow-starRows-treeDuration][imageCol][0]))+int(pgm_read_byte_near(&presents[presentRow2-starRows-treeDuration][imageCol][0])))>>1;
            g = (int(pgm_read_byte_near(&presents[imageRow-starRows-treeDuration][imageCol][1]))+int(pgm_read_byte_near(&presents[presentRow2-starRows-treeDuration][imageCol][1])))>>1;
            b = (int(pgm_read_byte_near(&presents[imageRow-starRows-treeDuration][imageCol][2]))+int(pgm_read_byte_near(&presents[presentRow2-starRows-treeDuration][imageCol][2])))>>1;
          }
          else if (imageRow >= starRows) {  // Draw tree lights
            snowValue = int(snow[y2][x+panel*numCols]);
            snowValue2 = int(snow[y2+imageRow2-imageRow][x+panel*numCols]);
//            snowValue2 = snowValue;
            r = (int(colors[snowValue][0])+int(colors[snowValue2][0]))>>1;
            g = (int(colors[snowValue][1])+int(colors[snowValue2][1]))>>1;
            b = (int(colors[snowValue][2])+int(colors[snowValue2][2]))>>1;
          }
          else if ((starCol>=0)&&(starCol<starCols)) {  // draw star
            starRow2 = min(imageRow2,starRows-1);
            r = (int(pgm_read_byte_near(&star[imageRow][starCol][0]))+int(pgm_read_byte_near(&star[starRow2][starCol][0])))>>1;
            g = (int(pgm_read_byte_near(&star[imageRow][starCol][1]))+int(pgm_read_byte_near(&star[starRow2][starCol][1])))>>1;
            b = (int(pgm_read_byte_near(&star[imageRow][starCol][2]))+int(pgm_read_byte_near(&star[starRow2][starCol][2])))>>1;
          }
          else {   // if you get here, you are outside the star image, draw black
            r = 0;
            g = 0;
            b = 0;
          }
          drawPixel(strip.Color(r, g, b)); // Draw pixel
          pixel++;
        }
        imageRow--;
        imageRow2--;
        y2++;
        for (x=numCols-1; x>=0; x--) {
          imageCol = imageStartCol + x;
          starCol = starStartCol + x;
          if ((imageRow < 0)||(imageRow >= starRows+treeDuration+presentRows)) {  // Draw black (star not present yet)
            r = 0;
            g = 0;
            b = 0;
          }
          else if (imageRow >= starRows+treeDuration) {  // draw presents
            presentRow2 = min(imageRow2,presentRows+starRows+treeDuration-1);
            r = (int(pgm_read_byte_near(&presents[imageRow-starRows-treeDuration][imageCol][0]))+int(pgm_read_byte_near(&presents[presentRow2-starRows-treeDuration][imageCol][0])))>>1;
            g = (int(pgm_read_byte_near(&presents[imageRow-starRows-treeDuration][imageCol][1]))+int(pgm_read_byte_near(&presents[presentRow2-starRows-treeDuration][imageCol][1])))>>1;
            b = (int(pgm_read_byte_near(&presents[imageRow-starRows-treeDuration][imageCol][2]))+int(pgm_read_byte_near(&presents[presentRow2-starRows-treeDuration][imageCol][2])))>>1;
          }
          else if (imageRow >= starRows) {  // Draw tree lights
            snowValue = int(snow[y2][x+panel*numCols]);
            snowValue2 = int(snow[y2+imageRow-imageRow2][x+panel*numCols]);
//
            snowValue2 = snowValue;
            r = (int(colors[snowValue][0])+int(colors[snowValue2][0]))>>1;
            g = (int(colors[snowValue][1])+int(colors[snowValue2][1]))>>1;
            b = (int(colors[snowValue][2])+int(colors[snowValue2][2]))>>1;
          }
          else if ((starCol>=0)&&(starCol<starCols)) {  // draw star
            starRow2 = min(imageRow2,starRows-1);
            r = (int(pgm_read_byte_near(&star[imageRow][starCol][0]))+int(pgm_read_byte_near(&star[starRow2][starCol][0])))>>1;
            g = (int(pgm_read_byte_near(&star[imageRow][starCol][1]))+int(pgm_read_byte_near(&star[starRow2][starCol][1])))>>1;
            b = (int(pgm_read_byte_near(&star[imageRow][starCol][2]))+int(pgm_read_byte_near(&star[starRow2][starCol][2])))>>1;
          }
          else {   // if you get here, you are outside the star image, draw black
            r = 0;
            g = 0;
            b = 0;
          }
          drawPixel(strip.Color(r, g, b)); // Draw image pixel
          pixel++;
        }
        imageRow--;
        imageRow2--;
      }
      strip.setPin(panelPin[panel]);
      strip.show();
    }
    delay(0);  // you can change this number to change the scrolling speed.  Smaller number will be faster scrolling.
  }

  flag++;
  if (flag>2) flag = 1;

}

static void drawPixel(uint32_t c) {
  strip.setPixelColor(pixel , c); // Draw new pixel
}
